@@ -10,6 +10,7 @@ package gcimporter
 import (
 	"bytes"
 	"fmt"
+	"go/build"
 	"go/constant"
 	"go/types"
 	"io/ioutil"
@@ -22,7 +23,6 @@ import (
 	"time"
 
 	"golang.org/x/tools/internal/testenv"
-	"golang.org/x/tools/internal/typeparams"
 )
 
 func TestMain(m *testing.M) {
@@ -32,20 +32,6 @@ func TestMain(m *testing.M) {
 
 // ----------------------------------------------------------------------------
 
-// skipSpecialPlatforms causes the test to be skipped for platforms where
-// builders (build.golang.org) don't have access to compiled packages for
-// import.
-func skipSpecialPlatforms(t *testing.T) {
-	switch platform := runtime.GOOS + "-" + runtime.GOARCH; platform {
-	case "nacl-amd64p32",
-		"nacl-386",
-		"nacl-arm",
-		"darwin-arm",
-		"darwin-arm64":
-		t.Skipf("no compiled packages available for import on %s", platform)
-	}
-}
-
 func needsCompiler(t *testing.T, compiler string) {
 	if runtime.Compiler == compiler {
 		return
@@ -67,7 +53,7 @@ func compile(t *testing.T, dirname, filename, outdirname string) string {
 	}
 	basename := filepath.Base(filename)
 	outname := filepath.Join(outdirname, basename[:len(basename)-2]+"o")
-	cmd := exec.Command("go", "tool", "compile", "-o", outname, filename)
+	cmd := exec.Command("go", "tool", "compile", "-p=p", "-o", outname, filename)
 	cmd.Dir = dirname
 	out, err := cmd.CombinedOutput()
 	if err != nil {
@@ -107,12 +93,6 @@ func testDir(t *testing.T, dir string, endTime time.Time) (nimports int) {
 			for _, ext := range pkgExts {
 				if strings.HasSuffix(f.Name(), ext) {
 					name := f.Name()[0 : len(f.Name())-len(ext)] // remove extension
-					if testenv.UsesGenerics(name) && !typeparams.Enabled {
-						// golang/go#48632: Skip generic packages when type parameters are
-						// not enabled, as a temporary measure to allow avoiding the new
-						// go/types API on Go 1.18.
-						continue
-					}
 					if testPath(t, filepath.Join(dir, name), dir) != nil {
 						nimports++
 					}
@@ -169,7 +149,9 @@ func TestImportTestdata(t *testing.T) {
 }
 
 func TestVersionHandling(t *testing.T) {
-	skipSpecialPlatforms(t) // we really only need to exclude nacl platforms, but this is fine
+	if debug {
+		t.Skip("TestVersionHandling panics in debug mode")
+	}
 
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
@@ -248,8 +230,6 @@ func TestVersionHandling(t *testing.T) {
 }
 
 func TestImportStdLib(t *testing.T) {
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -276,7 +256,7 @@ var importedObjectTests = []struct {
 	{"go/internal/gcimporter.FindPkg", "func FindPkg(path string, srcDir string) (filename string, id string)"},
 
 	// interfaces
-	{"context.Context", "type Context interface{Deadline() (deadline time.Time, ok bool); Done() <-chan struct{}; Err() error; Value(key interface{}) interface{}}"},
+	{"context.Context", "type Context interface{Deadline() (deadline time.Time, ok bool); Done() <-chan struct{}; Err() error; Value(key any) any}"},
 	{"crypto.Decrypter", "type Decrypter interface{Decrypt(rand io.Reader, msg []byte, opts DecrypterOpts) (plaintext []byte, err error); Public() PublicKey}"},
 	{"encoding.BinaryMarshaler", "type BinaryMarshaler interface{MarshalBinary() (data []byte, err error)}"},
 	{"io.Reader", "type Reader interface{Read(p []byte) (n int, err error)}"},
@@ -285,10 +265,20 @@ var importedObjectTests = []struct {
 	{"go/types.Type", "type Type interface{String() string; Underlying() Type}"},
 }
 
+// TODO(rsc): Delete this init func after x/tools no longer needs to test successfully with Go 1.17.
+func init() {
+	if build.Default.ReleaseTags[len(build.Default.ReleaseTags)-1] <= "go1.17" {
+		for i := range importedObjectTests {
+			if importedObjectTests[i].name == "context.Context" {
+				// Expand any to interface{}.
+				importedObjectTests[i].want = "type Context interface{Deadline() (deadline time.Time, ok bool); Done() <-chan struct{}; Err() error; Value(key interface{}) interface{}}"
+			}
+		}
+	}
+}
+
 func TestImportedTypes(t *testing.T) {
 	testenv.NeedsGo1Point(t, 11)
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -298,6 +288,12 @@ func TestImportedTypes(t *testing.T) {
 			continue // error reported elsewhere
 		}
 		got := types.ObjectString(obj, types.RelativeTo(obj.Pkg()))
+
+		// TODO(rsc): Delete this block once go.dev/cl/368254 lands.
+		if got != test.want && test.want == strings.ReplaceAll(got, "interface{}", "any") {
+			got = test.want
+		}
+
 		if got != test.want {
 			t.Errorf("%s: got %q; want %q", test.name, got, test.want)
 		}
@@ -310,8 +306,6 @@ func TestImportedTypes(t *testing.T) {
 
 func TestImportedConsts(t *testing.T) {
 	testenv.NeedsGo1Point(t, 11)
-	skipSpecialPlatforms(t)
-
 	tests := []struct {
 		name string
 		want constant.Kind
@@ -393,8 +387,6 @@ func verifyInterfaceMethodRecvs(t *testing.T, named *types.Named, level int) {
 }
 
 func TestIssue5815(t *testing.T) {
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -420,8 +412,6 @@ func TestIssue5815(t *testing.T) {
 
 // Smoke test to ensure that imported methods get the correct package.
 func TestCorrectMethodPackage(t *testing.T) {
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -441,8 +431,6 @@ func TestCorrectMethodPackage(t *testing.T) {
 }
 
 func TestIssue13566(t *testing.T) {
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -478,8 +466,6 @@ func TestIssue13566(t *testing.T) {
 }
 
 func TestIssue13898(t *testing.T) {
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -522,8 +508,6 @@ func TestIssue13898(t *testing.T) {
 }
 
 func TestIssue15517(t *testing.T) {
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -559,8 +543,6 @@ func TestIssue15517(t *testing.T) {
 }
 
 func TestIssue15920(t *testing.T) {
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -574,8 +556,6 @@ func TestIssue15920(t *testing.T) {
 }
 
 func TestIssue20046(t *testing.T) {
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -595,8 +575,6 @@ func TestIssue20046(t *testing.T) {
 
 func TestIssue25301(t *testing.T) {
 	testenv.NeedsGo1Point(t, 11)
-	skipSpecialPlatforms(t)
-
 	// This package only handles gc export data.
 	needsCompiler(t, "gc")
 
@@ -609,6 +587,37 @@ func TestIssue25301(t *testing.T) {
 	compileAndImportPkg(t, "issue25301")
 }
 
+func TestIssue51836(t *testing.T) {
+	testenv.NeedsGo1Point(t, 18) // requires generics
+
+	// This package only handles gc export data.
+	needsCompiler(t, "gc")
+
+	// On windows, we have to set the -D option for the compiler to avoid having a drive
+	// letter and an illegal ':' in the import path - just skip it (see also issue #3483).
+	if runtime.GOOS == "windows" {
+		t.Skip("avoid dealing with relative paths/drive letters on windows")
+	}
+
+	tmpdir := mktmpdir(t)
+	defer os.RemoveAll(tmpdir)
+	testoutdir := filepath.Join(tmpdir, "testdata")
+
+	dir := filepath.Join("testdata", "issue51836")
+	// Following the pattern of TestIssue13898, aa.go needs to be compiled from
+	// the output directory. We pass the full path to compile() so that we don't
+	// have to copy the file to that directory.
+	bpath, err := filepath.Abs(filepath.Join(dir, "aa.go"))
+	if err != nil {
+		t.Fatal(err)
+	}
+	compile(t, dir, "a.go", testoutdir)
+	compile(t, testoutdir, bpath, testoutdir)
+
+	// import must succeed (test for issue at hand)
+	_ = importPkg(t, "./testdata/aa", tmpdir)
+}
+
 func importPkg(t *testing.T, path, srcDir string) *types.Package {
 	pkg, err := Import(make(map[string]*types.Package), path, srcDir, nil)
 	if err != nil {
