@@ -13,6 +13,7 @@ package main // import "golang.org/x/tools/gopls"
 
 import (
 	"context"
+	"golang.org/x/tools/internal/analysisinternal"
 	"os"
 
 	"golang.org/x/tools/gopls/internal/hooks"
@@ -21,6 +22,10 @@ import (
 )
 
 func main() {
+	// In 1.18, diagnostics for Fuzz tests must not be used by cmd/vet.
+	// So the code for Fuzz tests diagnostics is guarded behind flag analysisinternal.DiagnoseFuzzTests
+	// Turn on analysisinternal.DiagnoseFuzzTests for gopls
+	analysisinternal.DiagnoseFuzzTests = true
 	ctx := context.Background()
 	tool.Main(ctx, cmd.New("gopls", "", nil, hooks.Options), os.Args[1:])
 }
