@@ -23,10 +23,10 @@ import (
 	"golang.org/x/tools/internal/lsp/diff"
 	"golang.org/x/tools/internal/lsp/diff/myers"
 	"golang.org/x/tools/internal/lsp/protocol"
+	"golang.org/x/tools/internal/lsp/safetoken"
 	"golang.org/x/tools/internal/lsp/source"
 	"golang.org/x/tools/internal/memoize"
 	"golang.org/x/tools/internal/span"
-	errors "golang.org/x/xerrors"
 )
 
 // parseKey uniquely identifies a parsed Go file.
@@ -72,27 +72,19 @@ func (s *snapshot) parseGoHandle(ctx context.Context, fh source.FileHandle, mode
 }
 
 func (pgh *parseGoHandle) String() string {
-	return pgh.File().URI().Filename()
-}
-
-func (pgh *parseGoHandle) File() source.FileHandle {
-	return pgh.file
-}
-
-func (pgh *parseGoHandle) Mode() source.ParseMode {
-	return pgh.mode
+	return pgh.file.URI().Filename()
 }
 
 func (s *snapshot) ParseGo(ctx context.Context, fh source.FileHandle, mode source.ParseMode) (*source.ParsedGoFile, error) {
-	pgh := s.parseGoHandle(ctx, fh, mode)
-	pgf, _, err := s.parseGo(ctx, pgh)
+	pgf, _, err := s.parseGo(ctx, fh, mode)
 	return pgf, err
 }
 
-func (s *snapshot) parseGo(ctx context.Context, pgh *parseGoHandle) (*source.ParsedGoFile, bool, error) {
-	if pgh.mode == source.ParseExported {
+func (s *snapshot) parseGo(ctx context.Context, fh source.FileHandle, mode source.ParseMode) (*source.ParsedGoFile, bool, error) {
+	if mode == source.ParseExported {
 		panic("only type checking should use Exported")
 	}
+	pgh := s.parseGoHandle(ctx, fh, mode)
 	d, err := pgh.handle.Get(ctx, s.generation, s)
 	if err != nil {
 		return nil, false, err
@@ -101,6 +93,22 @@ func (s *snapshot) parseGo(ctx context.Context, pgh *parseGoHandle) (*source.Par
 	return data.parsed, data.fixed, data.err
 }
 
+// cachedPGF returns the cached ParsedGoFile for the given ParseMode, if it
+// has already been computed. Otherwise, it returns nil.
+func (s *snapshot) cachedPGF(fh source.FileHandle, mode source.ParseMode) *source.ParsedGoFile {
+	key := parseKey{file: fh.FileIdentity(), mode: mode}
+	if pgh := s.getGoFile(key); pgh != nil {
+		cached := pgh.handle.Cached(s.generation)
+		if cached != nil {
+			cached := cached.(*parseGoData)
+			if cached.parsed != nil {
+				return cached.parsed
+			}
+		}
+	}
+	return nil
+}
+
 type astCacheKey struct {
 	pkg packageHandleKey
 	uri span.URI
@@ -249,7 +257,7 @@ func parseGo(ctx context.Context, fset *token.FileSet, fh source.FileHandle, mod
 
 	ext := filepath.Ext(fh.URI().Filename())
 	if ext != ".go" && ext != "" { // files generated by cgo have no extension
-		return &parseGoData{err: errors.Errorf("cannot parse non-Go file %s", fh.URI())}
+		return &parseGoData{err: fmt.Errorf("cannot parse non-Go file %s", fh.URI())}
 	}
 	src, err := fh.Read()
 	if err != nil {
@@ -323,9 +331,9 @@ func parseGo(ctx context.Context, fset *token.FileSet, fh source.FileHandle, mod
 			File: file,
 			Tok:  tok,
 			Mapper: &protocol.ColumnMapper{
-				URI:       fh.URI(),
-				Converter: span.NewTokenConverter(fset, tok),
-				Content:   src,
+				URI:     fh.URI(),
+				TokFile: tok,
+				Content: src,
 			},
 			ParseErr: parseErr,
 		},
@@ -686,7 +694,7 @@ func fixAST(ctx context.Context, n ast.Node, tok *token.File, src []byte) (fixed
 				// Recursively fix in our fixed node.
 				_ = fixAST(ctx, parent, tok, src)
 			} else {
-				err = errors.Errorf("unable to parse defer or go from *ast.BadStmt: %v", err)
+				err = fmt.Errorf("unable to parse defer or go from *ast.BadStmt: %v", err)
 			}
 			return false
 		case *ast.BadExpr:
@@ -756,7 +764,7 @@ func walkASTWithParent(n ast.Node, f func(n ast.Node, parent ast.Node) bool) {
 
 // fixSrc attempts to modify the file's source code to fix certain
 // syntax errors that leave the rest of the file unparsed.
-func fixSrc(f *ast.File, tok *token.File, src []byte) (newSrc []byte) {
+func fixSrc(f *ast.File, tf *token.File, src []byte) (newSrc []byte) {
 	walkASTWithParent(f, func(n, parent ast.Node) bool {
 		if newSrc != nil {
 			return false
@@ -764,9 +772,9 @@ func fixSrc(f *ast.File, tok *token.File, src []byte) (newSrc []byte) {
 
 		switch n := n.(type) {
 		case *ast.BlockStmt:
-			newSrc = fixMissingCurlies(f, n, parent, tok, src)
+			newSrc = fixMissingCurlies(f, n, parent, tf, src)
 		case *ast.SelectorExpr:
-			newSrc = fixDanglingSelector(n, tok, src)
+			newSrc = fixDanglingSelector(n, tf, src)
 		}
 
 		return newSrc == nil
@@ -778,16 +786,16 @@ func fixSrc(f *ast.File, tok *token.File, src []byte) (newSrc []byte) {
 // fixMissingCurlies adds in curly braces for block statements that
 // are missing curly braces. For example:
 //
-//   if foo
+//	if foo
 //
 // becomes
 //
-//   if foo {}
+//	if foo {}
 func fixMissingCurlies(f *ast.File, b *ast.BlockStmt, parent ast.Node, tok *token.File, src []byte) []byte {
 	// If the "{" is already in the source code, there isn't anything to
 	// fix since we aren't missing curlies.
 	if b.Lbrace.IsValid() {
-		braceOffset, err := source.Offset(tok, b.Lbrace)
+		braceOffset, err := safetoken.Offset(tok, b.Lbrace)
 		if err != nil {
 			return nil
 		}
@@ -842,7 +850,7 @@ func fixMissingCurlies(f *ast.File, b *ast.BlockStmt, parent ast.Node, tok *toke
 
 	var buf bytes.Buffer
 	buf.Grow(len(src) + 3)
-	offset, err := source.Offset(tok, insertPos)
+	offset, err := safetoken.Offset(tok, insertPos)
 	if err != nil {
 		return nil
 	}
@@ -875,23 +883,22 @@ func fixMissingCurlies(f *ast.File, b *ast.BlockStmt, parent ast.Node, tok *toke
 // "case" and "default" keywords as inside the switch statement. For
 // example:
 //
-//   switch {
-//   def<>
-//   }
+//	switch {
+//	def<>
+//	}
 //
 // gets parsed like:
 //
-//   switch {
-//   }
+//	switch {
+//	}
 //
 // Later we manually pull out the "def" token, but we need to detect
 // that our "<>" position is inside the switch block. To do that we
 // move the curly brace so it looks like:
 //
-//   switch {
-//
-//   }
+//	switch {
 //
+//	}
 func fixEmptySwitch(body *ast.BlockStmt, tok *token.File, src []byte) {
 	// We only care about empty switch statements.
 	if len(body.List) > 0 || !body.Rbrace.IsValid() {
@@ -900,7 +907,7 @@ func fixEmptySwitch(body *ast.BlockStmt, tok *token.File, src []byte) {
 
 	// If the right brace is actually in the source code at the
 	// specified position, don't mess with it.
-	braceOffset, err := source.Offset(tok, body.Rbrace)
+	braceOffset, err := safetoken.Offset(tok, body.Rbrace)
 	if err != nil {
 		return
 	}
@@ -921,16 +928,17 @@ func fixEmptySwitch(body *ast.BlockStmt, tok *token.File, src []byte) {
 // fixDanglingSelector inserts real "_" selector expressions in place
 // of phantom "_" selectors. For example:
 //
-// func _() {
-//   x.<>
-// }
+//	func _() {
+//		x.<>
+//	}
+//
 // var x struct { i int }
 //
 // To fix completion at "<>", we insert a real "_" after the "." so the
 // following declaration of "x" can be parsed and type checked
 // normally.
-func fixDanglingSelector(s *ast.SelectorExpr, tok *token.File, src []byte) []byte {
-	if !isPhantomUnderscore(s.Sel, tok, src) {
+func fixDanglingSelector(s *ast.SelectorExpr, tf *token.File, src []byte) []byte {
+	if !isPhantomUnderscore(s.Sel, tf, src) {
 		return nil
 	}
 
@@ -938,7 +946,7 @@ func fixDanglingSelector(s *ast.SelectorExpr, tok *token.File, src []byte) []byt
 		return nil
 	}
 
-	insertOffset, err := source.Offset(tok, s.X.End())
+	insertOffset, err := safetoken.Offset(tf, s.X.End())
 	if err != nil {
 		return nil
 	}
@@ -965,8 +973,8 @@ func fixDanglingSelector(s *ast.SelectorExpr, tok *token.File, src []byte) []byt
 // yields a "_" selector instead of "var" since "var" is a keyword.
 //
 // TODO(rfindley): should this constitute an ast 'fix'?
-func fixPhantomSelector(sel *ast.SelectorExpr, tok *token.File, src []byte) {
-	if !isPhantomUnderscore(sel.Sel, tok, src) {
+func fixPhantomSelector(sel *ast.SelectorExpr, tf *token.File, src []byte) {
+	if !isPhantomUnderscore(sel.Sel, tf, src) {
 		return
 	}
 
@@ -980,7 +988,7 @@ func fixPhantomSelector(sel *ast.SelectorExpr, tok *token.File, src []byte) {
 		return
 	}
 
-	maybeKeyword := readKeyword(sel.Sel.Pos(), tok, src)
+	maybeKeyword := readKeyword(sel.Sel.Pos(), tf, src)
 	if maybeKeyword == "" {
 		return
 	}
@@ -1001,7 +1009,7 @@ func isPhantomUnderscore(id *ast.Ident, tok *token.File, src []byte) bool {
 
 	// Phantom underscore means the underscore is not actually in the
 	// program text.
-	offset, err := source.Offset(tok, id.Pos())
+	offset, err := safetoken.Offset(tok, id.Pos())
 	if err != nil {
 		return false
 	}
@@ -1021,11 +1029,11 @@ func fixInitStmt(bad *ast.BadExpr, parent ast.Node, tok *token.File, src []byte)
 	}
 
 	// Try to extract a statement from the BadExpr.
-	start, err := source.Offset(tok, bad.Pos())
+	start, err := safetoken.Offset(tok, bad.Pos())
 	if err != nil {
 		return
 	}
-	end, err := source.Offset(tok, bad.End()-1)
+	end, err := safetoken.Offset(tok, bad.End()-1)
 	if err != nil {
 		return
 	}
@@ -1069,7 +1077,7 @@ func fixInitStmt(bad *ast.BadExpr, parent ast.Node, tok *token.File, src []byte)
 // readKeyword reads the keyword starting at pos, if any.
 func readKeyword(pos token.Pos, tok *token.File, src []byte) string {
 	var kwBytes []byte
-	offset, err := source.Offset(tok, pos)
+	offset, err := safetoken.Offset(tok, pos)
 	if err != nil {
 		return ""
 	}
@@ -1110,11 +1118,11 @@ func fixArrayType(bad *ast.BadExpr, parent ast.Node, tok *token.File, src []byte
 	// Avoid doing tok.Offset(to) since that panics if badExpr ends at EOF.
 	// It also panics if the position is not in the range of the file, and
 	// badExprs may not necessarily have good positions, so check first.
-	fromOffset, err := source.Offset(tok, from)
+	fromOffset, err := safetoken.Offset(tok, from)
 	if err != nil {
 		return false
 	}
-	toOffset, err := source.Offset(tok, to-1)
+	toOffset, err := safetoken.Offset(tok, to-1)
 	if err != nil {
 		return false
 	}
@@ -1271,7 +1279,7 @@ FindTo:
 		}
 	}
 
-	fromOffset, err := source.Offset(tok, from)
+	fromOffset, err := safetoken.Offset(tok, from)
 	if err != nil {
 		return false
 	}
@@ -1279,7 +1287,7 @@ FindTo:
 		return false
 	}
 
-	toOffset, err := source.Offset(tok, to)
+	toOffset, err := safetoken.Offset(tok, to)
 	if err != nil {
 		return false
 	}
@@ -1339,17 +1347,17 @@ func parseStmt(pos token.Pos, src []byte) (ast.Stmt, error) {
 	// best-effort behavior, whereas ParseExpr fails hard on any error.
 	fakeFile, err := parser.ParseFile(token.NewFileSet(), "", fileSrc, 0)
 	if fakeFile == nil {
-		return nil, errors.Errorf("error reading fake file source: %v", err)
+		return nil, fmt.Errorf("error reading fake file source: %v", err)
 	}
 
 	// Extract our expression node from inside the fake file.
 	if len(fakeFile.Decls) == 0 {
-		return nil, errors.Errorf("error parsing fake file: %v", err)
+		return nil, fmt.Errorf("error parsing fake file: %v", err)
 	}
 
 	fakeDecl, _ := fakeFile.Decls[0].(*ast.FuncDecl)
 	if fakeDecl == nil || len(fakeDecl.Body.List) == 0 {
-		return nil, errors.Errorf("no statement in %s: %v", src, err)
+		return nil, fmt.Errorf("no statement in %s: %v", src, err)
 	}
 
 	stmt := fakeDecl.Body.List[0]
@@ -1371,7 +1379,7 @@ func parseExpr(pos token.Pos, src []byte) (ast.Expr, error) {
 
 	exprStmt, ok := stmt.(*ast.ExprStmt)
 	if !ok {
-		return nil, errors.Errorf("no expr in %s: %v", src, err)
+		return nil, fmt.Errorf("no expr in %s: %v", src, err)
 	}
 
 	return exprStmt.X, nil
