@@ -18,7 +18,6 @@ import (
 	"golang.org/x/tools/internal/lsp/source"
 	"golang.org/x/tools/internal/span"
 	"golang.org/x/tools/internal/xcontext"
-	errors "golang.org/x/xerrors"
 )
 
 type Session struct {
@@ -155,7 +154,7 @@ func (s *Session) Cache() interface{} {
 	return s.cache
 }
 
-func (s *Session) NewView(ctx context.Context, name string, folder, tempWorkspace span.URI, options *source.Options) (source.View, source.Snapshot, func(), error) {
+func (s *Session) NewView(ctx context.Context, name string, folder span.URI, options *source.Options) (source.View, source.Snapshot, func(), error) {
 	s.viewMu.Lock()
 	defer s.viewMu.Unlock()
 	for _, view := range s.views {
@@ -163,7 +162,7 @@ func (s *Session) NewView(ctx context.Context, name string, folder, tempWorkspac
 			return nil, nil, nil, source.ErrViewExists
 		}
 	}
-	view, snapshot, release, err := s.createView(ctx, name, folder, tempWorkspace, options, 0)
+	view, snapshot, release, err := s.createView(ctx, name, folder, options, 0)
 	if err != nil {
 		return nil, nil, func() {}, err
 	}
@@ -173,7 +172,7 @@ func (s *Session) NewView(ctx context.Context, name string, folder, tempWorkspac
 	return view, snapshot, release, nil
 }
 
-func (s *Session) createView(ctx context.Context, name string, folder, tempWorkspace span.URI, options *source.Options, snapshotID uint64) (*View, *snapshot, func(), error) {
+func (s *Session) createView(ctx context.Context, name string, folder span.URI, options *source.Options, snapshotID uint64) (*View, *snapshot, func(), error) {
 	index := atomic.AddInt64(&viewIndex, 1)
 
 	if s.cache.options != nil {
@@ -218,7 +217,6 @@ func (s *Session) createView(ctx context.Context, name string, folder, tempWorks
 		filesByBase:          map[string][]*fileBase{},
 		rootURI:              root,
 		workspaceInformation: *ws,
-		tempWorkspace:        tempWorkspace,
 	}
 	v.importsState = &importsState{
 		ctx: backgroundCtx,
@@ -234,16 +232,15 @@ func (s *Session) createView(ctx context.Context, name string, folder, tempWorks
 		initializeOnce:    &sync.Once{},
 		generation:        s.cache.store.Generation(generationName(v, 0)),
 		packages:          make(map[packageKey]*packageHandle),
-		ids:               make(map[span.URI][]PackageID),
-		metadata:          make(map[PackageID]*KnownMetadata),
+		meta:              NewMetadataGraph(),
 		files:             make(map[span.URI]source.VersionedFileHandle),
 		goFiles:           make(map[parseKey]*parseGoHandle),
 		symbols:           make(map[span.URI]*symbolHandle),
-		importedBy:        make(map[PackageID][]PackageID),
 		actions:           make(map[actionKey]*actionHandle),
 		workspacePackages: make(map[PackageID]PackagePath),
 		unloadableFiles:   make(map[span.URI]struct{}),
 		parseModHandles:   make(map[span.URI]*parseModHandle),
+		parseWorkHandles:  make(map[span.URI]*parseWorkHandle),
 		modTidyHandles:    make(map[span.URI]*modTidyHandle),
 		modWhyHandles:     make(map[span.URI]*modWhyHandle),
 		workspace:         workspace,
@@ -257,11 +254,6 @@ func (s *Session) createView(ctx context.Context, name string, folder, tempWorks
 	go func() {
 		defer release()
 		snapshot.initialize(initCtx, true)
-		// Ensure that the view workspace is written at least once following
-		// initialization.
-		if err := v.updateWorkspace(initCtx); err != nil {
-			event.Error(ctx, "copying workspace dir", err)
-		}
 	}()
 	return v, snapshot, snapshot.generation.Acquire(), nil
 }
@@ -381,7 +373,7 @@ func (s *Session) updateView(ctx context.Context, view *View, options *source.Op
 		return nil, err
 	}
 
-	v, _, release, err := s.createView(ctx, view.name, view.folder, view.tempWorkspace, options, snapshotID)
+	v, _, release, err := s.createView(ctx, view.name, view.folder, options, snapshotID)
 	release()
 
 	if err != nil {
@@ -409,7 +401,7 @@ func (s *Session) dropView(ctx context.Context, v *View) (int, error) {
 			return i, nil
 		}
 	}
-	return -1, errors.Errorf("view %s for %v not found", v.Name(), v.Folder())
+	return -1, fmt.Errorf("view %s for %v not found", v.Name(), v.Folder())
 }
 
 func (s *Session) ModifyFiles(ctx context.Context, changes []source.FileModification) error {
@@ -619,7 +611,7 @@ func (s *Session) updateOverlays(ctx context.Context, changes []source.FileModif
 			kind = source.FileKindForLang(c.LanguageID)
 		default:
 			if !ok {
-				return nil, errors.Errorf("updateOverlays: modifying unopened overlay %v", c.URI)
+				return nil, fmt.Errorf("updateOverlays: modifying unopened overlay %v", c.URI)
 			}
 			kind = o.kind
 		}
@@ -653,10 +645,10 @@ func (s *Session) updateOverlays(ctx context.Context, changes []source.FileModif
 		case source.Save:
 			// Make sure the version and content (if present) is the same.
 			if false && o.version != version { // Client no longer sends the version
-				return nil, errors.Errorf("updateOverlays: saving %s at version %v, currently at %v", c.URI, c.Version, o.version)
+				return nil, fmt.Errorf("updateOverlays: saving %s at version %v, currently at %v", c.URI, c.Version, o.version)
 			}
 			if c.Text != nil && o.hash != hash {
-				return nil, errors.Errorf("updateOverlays: overlay %s changed on save", c.URI)
+				return nil, fmt.Errorf("updateOverlays: overlay %s changed on save", c.URI)
 			}
 			sameContentOnDisk = true
 		default:
@@ -681,10 +673,10 @@ func (s *Session) updateOverlays(ctx context.Context, changes []source.FileModif
 		if c.Action == source.Open {
 			view, err := s.ViewOf(o.uri)
 			if err != nil {
-				return nil, errors.Errorf("updateOverlays: finding view for %s: %v", o.uri, err)
+				return nil, fmt.Errorf("updateOverlays: finding view for %s: %v", o.uri, err)
 			}
 			if kind := view.FileKind(o); kind == source.UnknownKind {
-				return nil, errors.Errorf("updateOverlays: unknown file kind for %s", o.uri)
+				return nil, fmt.Errorf("updateOverlays: unknown file kind for %s", o.uri)
 			}
 		}
 
