@@ -17,58 +17,77 @@ import (
 	"strconv"
 	"strings"
 
+	"golang.org/x/mod/modfile"
+	"golang.org/x/tools/internal/lsp/bug"
 	"golang.org/x/tools/internal/lsp/protocol"
 	"golang.org/x/tools/internal/span"
-	errors "golang.org/x/xerrors"
 )
 
 // MappedRange provides mapped protocol.Range for a span.Range, accounting for
 // UTF-16 code points.
 type MappedRange struct {
-	spanRange span.Range
-	m         *protocol.ColumnMapper
-
-	// protocolRange is the result of converting the spanRange using the mapper.
-	// It is computed on-demand.
-	protocolRange *protocol.Range
+	spanRange span.Range             // the range in the compiled source (package.CompiledGoFiles)
+	m         *protocol.ColumnMapper // a mapper of the edited source (package.GoFiles)
 }
 
 // NewMappedRange returns a MappedRange for the given start and end token.Pos.
+//
+// By convention, start and end are assumed to be positions in the compiled (==
+// type checked) source, whereas the column mapper m maps positions in the
+// user-edited source. Note that these may not be the same, as when using CGo:
+// CompiledGoFiles contains generated files, whose positions (via
+// token.File.Position) point to locations in the edited file -- the file
+// containing `import "C"`.
 func NewMappedRange(fset *token.FileSet, m *protocol.ColumnMapper, start, end token.Pos) MappedRange {
+	if tf := fset.File(start); tf == nil {
+		bug.Report("nil file", nil)
+	} else {
+		mapped := m.TokFile.Name()
+		adjusted := tf.PositionFor(start, true) // adjusted position
+		if adjusted.Filename != mapped {
+			bug.Reportf("mapped file %q does not match start position file %q", mapped, adjusted.Filename)
+		}
+	}
 	return MappedRange{
-		spanRange: span.Range{
-			FileSet:   fset,
-			Start:     start,
-			End:       end,
-			Converter: m.Converter,
-		},
-		m: m,
+		spanRange: span.NewRange(fset, start, end),
+		m:         m,
 	}
 }
 
+// Range returns the LSP range in the edited source.
+//
+// See the documentation of NewMappedRange for information on edited vs
+// compiled source.
 func (s MappedRange) Range() (protocol.Range, error) {
-	if s.protocolRange == nil {
-		spn, err := s.spanRange.Span()
-		if err != nil {
-			return protocol.Range{}, err
-		}
-		prng, err := s.m.Range(spn)
-		if err != nil {
-			return protocol.Range{}, err
-		}
-		s.protocolRange = &prng
+	if s.m == nil {
+		return protocol.Range{}, bug.Errorf("invalid range")
 	}
-	return *s.protocolRange, nil
+	spn, err := s.Span()
+	if err != nil {
+		return protocol.Range{}, err
+	}
+	return s.m.Range(spn)
 }
 
+// Span returns the span corresponding to the mapped range in the edited
+// source.
+//
+// See the documentation of NewMappedRange for information on edited vs
+// compiled source.
 func (s MappedRange) Span() (span.Span, error) {
-	return s.spanRange.Span()
-}
-
-func (s MappedRange) SpanRange() span.Range {
-	return s.spanRange
+	// In the past, some code-paths have relied on Span returning an error if s
+	// is the zero value (i.e. s.m is nil). But this should be treated as a bug:
+	// observe that s.URI() would panic in this case.
+	if s.m == nil {
+		return span.Span{}, bug.Errorf("invalid range")
+	}
+	return span.FileSpan(s.spanRange.TokFile, s.m.TokFile, s.spanRange.Start, s.spanRange.End)
 }
 
+// URI returns the URI of the edited file.
+//
+// See the documentation of NewMappedRange for information on edited vs
+// compiled source.
 func (s MappedRange) URI() span.URI {
 	return s.m.URI
 }
@@ -94,15 +113,11 @@ func IsGenerated(ctx context.Context, snapshot Snapshot, uri span.URI) bool {
 	if err != nil {
 		return false
 	}
-	tok := snapshot.FileSet().File(pgf.File.Pos())
-	if tok == nil {
-		return false
-	}
 	for _, commentGroup := range pgf.File.Comments {
 		for _, comment := range commentGroup.List {
 			if matched := generatedRx.MatchString(comment.Text); matched {
 				// Check if comment is at the beginning of the line in source.
-				if pos := tok.Position(comment.Slash); pos.Column == 1 {
+				if pgf.Tok.Position(comment.Slash).Column == 1 {
 					return true
 				}
 			}
@@ -149,15 +164,16 @@ func posToMappedRange(snapshot Snapshot, pkg Package, pos, end token.Pos) (Mappe
 		return MappedRange{}, err
 	}
 	if !pos.IsValid() {
-		return MappedRange{}, errors.Errorf("invalid position for %v", pos)
+		return MappedRange{}, fmt.Errorf("invalid position for %v", pos)
 	}
 	if !end.IsValid() {
-		return MappedRange{}, errors.Errorf("invalid position for %v", end)
+		return MappedRange{}, fmt.Errorf("invalid position for %v", end)
 	}
 	return NewMappedRange(snapshot.FileSet(), pgf.Mapper, pos, end), nil
 }
 
 // Matches cgo generated comment as well as the proposed standard:
+//
 //	https://golang.org/s/generatedcode
 var generatedRx = regexp.MustCompile(`// .*DO NOT EDIT\.?`)
 
@@ -173,30 +189,13 @@ func FileKindForLang(langID string) FileKind {
 		return Sum
 	case "tmpl", "gotmpl":
 		return Tmpl
+	case "go.work":
+		return Work
 	default:
 		return UnknownKind
 	}
 }
 
-func DetectLanguage(langID, filename string) FileKind {
-	// use the langID if the client sent it
-	if kind := FileKindForLang(langID); kind != UnknownKind {
-		return kind
-	}
-	// Detect the language based on the file extension.
-	switch ext := filepath.Ext(filename); ext {
-	case ".mod":
-		return Mod
-	case ".sum":
-		return Sum
-	case ".go":
-		return Go
-	default:
-		// (for instance, before go1.15 cgo files had no extension)
-		return Go
-	}
-}
-
 func (k FileKind) String() string {
 	switch k {
 	case Go:
@@ -207,6 +206,8 @@ func (k FileKind) String() string {
 		return "go.sum"
 	case Tmpl:
 		return "tmpl"
+	case Work:
+		return "go.work"
 	default:
 		return fmt.Sprintf("unk%d", k)
 	}
@@ -277,13 +278,16 @@ func CompareDiagnostic(a, b *Diagnostic) int {
 	if a.Source < b.Source {
 		return -1
 	}
+	if a.Source > b.Source {
+		return +1
+	}
 	if a.Message < b.Message {
 		return -1
 	}
-	if a.Message == b.Message {
-		return 0
+	if a.Message > b.Message {
+		return +1
 	}
-	return 1
+	return 0
 }
 
 // FindPackageFromPos finds the first package containing pos in its
@@ -291,7 +295,7 @@ func CompareDiagnostic(a, b *Diagnostic) int {
 func FindPackageFromPos(ctx context.Context, snapshot Snapshot, pos token.Pos) (Package, error) {
 	tok := snapshot.FileSet().File(pos)
 	if tok == nil {
-		return nil, errors.Errorf("no file for pos %v", pos)
+		return nil, fmt.Errorf("no file for pos %v", pos)
 	}
 	uri := span.URIFromPath(tok.Name())
 	pkgs, err := snapshot.PackagesForFile(ctx, uri, TypecheckAll, true)
@@ -312,7 +316,7 @@ func FindPackageFromPos(ctx context.Context, snapshot Snapshot, pos token.Pos) (
 		}
 		return pkg, nil
 	}
-	return nil, errors.Errorf("no package for given file position")
+	return nil, fmt.Errorf("no package for given file position")
 }
 
 // findFileInDeps finds uri in pkg or its dependencies.
@@ -334,7 +338,7 @@ func findFileInDeps(pkg Package, uri span.URI) (*ParsedGoFile, Package, error) {
 			}
 		}
 	}
-	return nil, nil, errors.Errorf("no file for %s in package %s", uri, pkg.ID())
+	return nil, nil, fmt.Errorf("no file for %s in package %s", uri, pkg.ID())
 }
 
 // ImportPath returns the unquoted import path of s,
@@ -555,17 +559,22 @@ func IsCommandLineArguments(s string) bool {
 	return strings.Contains(s, "command-line-arguments")
 }
 
-// Offset returns tok.Offset(pos), but it also checks that the pos is in range
-// for the given file.
-func Offset(tok *token.File, pos token.Pos) (int, error) {
-	if !InRange(tok, pos) {
-		return -1, fmt.Errorf("pos %v is not in range for file [%v:%v)", pos, tok.Base(), tok.Base()+tok.Size())
-	}
-	return tok.Offset(pos), nil
+// LineToRange creates a Range spanning start and end.
+func LineToRange(m *protocol.ColumnMapper, uri span.URI, start, end modfile.Position) (protocol.Range, error) {
+	return ByteOffsetsToRange(m, uri, start.Byte, end.Byte)
 }
 
-// InRange reports whether the given position is in the given token.File.
-func InRange(tok *token.File, pos token.Pos) bool {
-	size := tok.Pos(tok.Size())
-	return int(pos) >= tok.Base() && pos <= size
+// ByteOffsetsToRange creates a range spanning start and end.
+func ByteOffsetsToRange(m *protocol.ColumnMapper, uri span.URI, start, end int) (protocol.Range, error) {
+	line, col, err := span.ToPosition(m.TokFile, start)
+	if err != nil {
+		return protocol.Range{}, err
+	}
+	s := span.NewPoint(line, col, start)
+	line, col, err = span.ToPosition(m.TokFile, end)
+	if err != nil {
+		return protocol.Range{}, err
+	}
+	e := span.NewPoint(line, col, end)
+	return m.Range(span.New(uri, s, e))
 }
