@@ -252,6 +252,7 @@ func DefaultOptions(o *source.Options) {
 			protocol.SourceOrganizeImports: true,
 		},
 		source.Sum:  {},
+		source.Work: {},
 		source.Tmpl: {},
 	}
 	o.UserOptions.Codelenses[string(command.Test)] = true
@@ -996,12 +997,7 @@ func (data *Data) Mapper(uri span.URI) (*protocol.ColumnMapper, error) {
 		if err != nil {
 			return nil, err
 		}
-		converter := span.NewContentConverter(uri.Filename(), content)
-		data.mappers[uri] = &protocol.ColumnMapper{
-			URI:       uri,
-			Converter: converter,
-			Content:   content,
-		}
+		data.mappers[uri] = protocol.NewColumnMapper(uri, content)
 	}
 	return data.mappers[uri], nil
 }
